#include <stdio.h>		/* For NULL ---- ick. */
#include <malloc.h>

#include "audio.h"
#include "convert.h"

/*
 * xplay --- X sound player.
 *
 * This program is copyright 1994 by Robert S. Thau.  It is offered as
 * is, and no warranty is offered of suitability or fitness for any
 * particular purpose; use at your own risk.  Contents may have settled
 * during shipment.  The program may be redistributed for noncommercial
 * purposes only, so long as all files contain a prominent notice of when
 * they were last changed, and who changed them.
 */

extern unsigned char ulaw_from_short[];
extern short short_from_ulaw[];

#ifndef __GCC__
#define inline
#endif

/*
 * File-format independant conversion routines ...
 *
 * Rate conversion --- we use linear interpolation.  Sox has this really
 * fancy hack involving the lcm of the sampling rates, which is neat except
 * for two things:
 *
 * 1) it goes overboard being exact about an approximation...
 * 2) it assumes that the sampling rates are integers.  Believe it or not,
 *    there are samples on the net which claim rates like 11085.3 Hz.
 *
 * We just keep track of the relative skew between the pre- and post-
 * conversion sample streams in nanoseconds.  (Note that it is important
 * to use the skew and not try to track absolute times, because otherwise
 * there will be overflow).
 *
 * Note that nsec_skew is kept as a float even though its value is always
 * an (exact) integer, to save conversion overhead when we do eventually
 * want to make fractions out of it...
 *
 * The data structures we use to keep track of this are a little more
 * overwrought than they actually need to be --- resetting the skew to
 * zero once every few hundred samples would probably be all right.
 * However, this is intended to be a hook for routines to process
 * compressed audio formats...
 */

struct conv_state
{
  struct format *dst;
  struct format *src;
  
  int nsec_per_src_sample;
  int nsec_per_dst_sample;
  
  int last_left;		/* Last source sample converted */
  int last_right;
  
  float nsec_skew;		/* How long we are past that in dst stream */
};

char *
check_for_conversion (struct format *src)
{
  /* This makes sure that the format we have is one that the
   * conversion routine below can handle, and fills in the
   * dst structure with details of the format which we will
   * be converting to.
   */

  char *errmsg = malloc(100);
  
  if (errmsg == NULL)
  {
    return "Couldn't malloc";
  }

  if (src->nbits != 8 && src->nbits != 16)
  {
    sprintf (errmsg, "Can't handle %d-bit samples", src->nbits);
    return errmsg;
  }
  
  if (src->channels != 1 && src->channels != 2)
  {
    sprintf (errmsg, "Can't handle data with %d channels", src->channels);
    return errmsg;
  }

  if (src->coding == CODE_ALAW
      && (src->channels != 1 || src->rate != 8000))
  {
    sprintf (errmsg,
	     "Sample is ALAW; can only handle that at 8 kHz mono;"
	     "but got %d channels at %d Hz",
	     src->channels, src->rate);
    return errmsg;
  }
  
  if (src->nbits == 16 && src->coding != CODE_LINEAR
      && src->coding != CODE_LINEAR_UNSIGNED)
  {
    strcpy (errmsg, "Can't handle 16-bit nonlinear data");
    return errmsg;
  }

  free(errmsg);
  return NULL;
}
  
void *
grok_for_conversion (struct format *dst, struct format *src, int *devspec)
{
  float rate_conversion;
  
  struct conv_state *state =
    (struct conv_state *)malloc(sizeof(struct conv_state));

  if (state == NULL) return NULL; /* Nasty error; caller checks. */
  
  /* We always make nonlinear output if we get nonlinear input;
   * otherwise, take device's preferred encoding.
   */

  if (src->coding == CODE_ULAW || src->coding == CODE_ALAW)
  {
    dst->coding = src->coding;
  }
  else dst->coding = devspec[0];

  /* Output stereo iff input is stereo and device can handle it */

  if (src->channels == 2 && devspec[1] == 2) dst->channels = 2;
  else dst->channels = 1;
  
  dst->nbits = devspec[2];

  if (src->coding != CODE_LINEAR)
    dst->rate = 8000;
  else
  {
    /* Adapt to sampling rate ... */

    int *ratep = &devspec[3];

    while (ratep[1] > 0)
      if (ratep[0] > src->rate)
	break;
      else
	++ratep;

    dst->rate = ratep[0];
  }
  
  /* Always output native byte order ... */

  {
    short kludge = 1;
    dst->little_endian = *(char*)&kludge;
  }
  
  rate_conversion = ((float)src->rate) / ((float)dst->rate);
  dst->num_samples = src->num_samples / rate_conversion;
  
  state->src = src;
  state->dst = dst;
  state->nsec_per_src_sample = 1.0e9 / src->rate;
  state->nsec_per_dst_sample = 1.0e9 / dst->rate;
  state->last_left = 0;
  state->last_right = 0;
  state->nsec_skew = 0;
  
  return (void*)state;
}

inline int get_sample(struct format *fmt, char **ptr)
{
  int next_val = *(*ptr)++;
    
  /* Note we don't worry about ALAW here, as it's special-cased
   * in the mainline of convert_samples...
   */

  if (fmt->coding == CODE_ULAW)
  {
    next_val = short_from_ulaw[(unsigned char)next_val];
  }
  else if (fmt->little_endian && fmt->nbits > 8)
  {
    int lobyte = next_val & 0xff;
    
    next_val = (*(*ptr)++) << 8 | lobyte;
  }
  else
  {
    next_val <<= 8;

    if (fmt->nbits > 8)
    {
      next_val |= *(unsigned char *)(*ptr);
      (*ptr)++;
    }
  }

  if (fmt->coding == CODE_LINEAR_UNSIGNED)
  {
    next_val &= 0xffff;
    next_val -= 32768;
  }
  
  return next_val;
}

inline void discard_sample(struct format *fmt, char **ptr)
{
  (*ptr) ++;
  if (fmt->nbits > 8) (*ptr) ++;
}
	
int sample_size (struct format *fmt)
{
  return fmt->channels * (fmt->nbits / 8);
}

/*
 * Main conversion routine.
 */

void
convert_samples (char **h_dest, char **h_src,
		 void *astate, int dst_nbytes, int src_nbytes)
{
  char *dest = *h_dest, *src = *h_src;
  struct conv_state *state = (struct conv_state *)astate;
  
  int last_left = state->last_left;
  int last_right = state->last_right;
  
  int next_left = 0, next_right = 0;
    
  float nsec_skew = state->nsec_skew;
  
  float src_nsec = state->nsec_per_src_sample;
  float dst_nsec = state->nsec_per_dst_sample;
  
  struct format *src_fmt = state->src;
  struct format *dst_fmt = state->dst;
  
  int src_chan = src_fmt->channels;
  int src_enc = src_fmt->coding;

  int dst_chan = dst_fmt->channels;
  int dst_enc = dst_fmt->coding;
  
  char *dstp = dest, *srcp = src;
  char *dst_end, *src_end;

  int ssize = sample_size (src_fmt);

  /* First, check for a special case ... 8-bit ALAW, which
   * the chip understands directly, and we don't.
   */

  if (src_enc == CODE_ALAW)
  {
    int nsamples = dst_nbytes;

    if (src_nbytes < dst_nbytes) nsamples = src_nbytes;
    
    memcpy (dest, src, nsamples);
    
    *h_dest += nsamples; *h_src += nsamples;
    return;
  }

  /* General case
   *
   * Compute limits of the iteration --- note that for src, we
   * are careful about not trying to convert a sample which
   * is only partially available...
   *
   * The code "knows" this is not an issue for dst, since dst
   * goes into buffers whose sizes are always sufficiently large
   * powers of two.  That's good, because if that situation ever
   * did arise, for dst, fill_buffer in play.c would hang forever.
   * Ugly, ugly, ugly...
   */

  dst_end = dstp + dst_nbytes;
  src_end = srcp + ssize * (src_nbytes / ssize); 
  
  /* Do it */
  
  while (srcp < src_end && dstp < dst_end)
  {
    /* Want to compute value for points which are nsec_skew
     * microseconds beyond last_sample... may have to advance
     * in the source stream even to get there.
     */
    
    while (srcp < src_end && nsec_skew > src_nsec + src_nsec)
    {
      discard_sample (src_fmt, &srcp);
      if (src_chan == 2) discard_sample (src_fmt, &srcp);
      nsec_skew -= src_nsec;
    }

    if (srcp < src_end && nsec_skew > src_nsec)
    {
      last_left = get_sample (src_fmt, &srcp);
      if (src_chan == 2) last_right = get_sample (src_fmt, &srcp);
      
      nsec_skew -= src_nsec;
    }

    if (srcp < src_end)
    {
      /* Now converting for dest points which are between last_val
       * and the next sample.  Get that sample.
       */
      
      next_left = get_sample (src_fmt, &srcp);
      if (src_chan == 2) next_right = get_sample (src_fmt, &srcp);

      /* Compute values by linear interpolation between the two. */
    
      while (nsec_skew < src_nsec && srcp < src_end && dstp < dst_end)
      {
	float cnext = nsec_skew / src_nsec;

	int left_val = cnext * next_left + (1.0 - cnext) * last_left;
	int right_val;
	
	if (src_chan == 2)
	{
	  /* Stereo input... */
	  right_val = cnext * next_right + (1.0 - cnext) * last_right;
	}
	
	if (dst_chan == 1)
	{
	  if (src_chan == 2) left_val = (left_val + right_val) >> 1;
	
	  if (dst_enc == CODE_ULAW)
	    *dstp++ = ulaw_from_short[32768 + left_val];
	  else			/* Linear, must be 16 bit */
	    *((short*)dstp)++ = left_val;
	}
	else
	{
	  /* grok_for_conversion assures that we never try to convert
	   * mono to stereo, but what the hell...
	   *
	   * What's more serious is that we don't check the output encoding.
	   */
	  
	  if (src_chan == 1) right_val = left_val;
	
	  *((short*)dstp)++ = left_val;
	  *((short*)dstp)++ = right_val;
	}
      
	nsec_skew += dst_nsec;
      }

      /* Gone beyond last_val.  Remember what we read out... */

      last_left = next_left;
      last_right = next_right;
      nsec_skew -= src_nsec;
    }
  }

  /* Save state for the next round */
  
  state->nsec_skew = nsec_skew;
  state->last_left = last_left;
  state->last_right = last_right;
  
  *h_dest = dstp;
  *h_src = srcp;
}
