#include <stdio.h>
#include <memory.h>
#include <malloc.h>

#include "audio.h"

/*
 * xplay --- X sound player.
 *
 * This program is copyright 1994 by Robert S. Thau.  It is offered as
 * is, and no warranty is offered of suitability or fitness for any
 * particular purpose; use at your own risk.  Contents may have settled
 * during shipment.  The program may be redistributed for noncommercial
 * purposes only, so long as all files contain a prominent notice of when
 * they were last changed, and who changed them.
 */

typedef unsigned long uint_32;
#define ROUND_UP(x) ((((int)x)&1) ? (x) + 1 : (x))

/* These macros side-effect the args --- beware!
 *
 * Note also that they don't take advantage of the fact that the Sparc
 * has the right byte-sex, because the data may very well have the
 * wrong alignment.  If you've gotta go byte-by-byte anyway...
 *
 * But the function which reads "extendeds" *does* know the order
 * of bytes in a double.
 *
 * Note also that you don't get all the bytes of a value which is
 * stored in the file as an 'ext'.
 */

typedef unsigned char *iff_handle;

#define IFF_ID(a,b,c,d) (((a)<<24) + ((b)<<16) + ((c)<<8) + (d))

#define IFF_get_char(ptr) (*(ptr)++)
#define IFF_get_short(ptr) ((ptr) += 2, ((ptr)[-2] << 8) + (ptr)[-1])
#define IFF_get_long(ptr) ((ptr) += 4, \
			   IFF_ID((ptr)[-4], (ptr)[-3], (ptr)[-2], (ptr)[-1]))

static double IFF_read_extended (iff_handle);
				  
#define IFF_get_ext(ptr) ((ptr) += 10, IFF_read_extended((ptr) - 10))

/* Chunk types we know about ... */

#define IFF_Form IFF_ID('F', 'O', 'R', 'M')

#define IFF_AIFF IFF_ID('A', 'I', 'F', 'F')
#define IFF_AIFC IFF_ID('A', 'I', 'F', 'C')

#define AIFF_COMM IFF_ID('C', 'O', 'M', 'M')
#define AIFF_SSND IFF_ID('S', 'S', 'N', 'D')
			  
/* Handling chunks ... */

#define IFF_chunk_type(ptr) (IFF_ID((ptr)[0], (ptr)[1], (ptr)[2], (ptr)[3]))
#define IFF_chunk_len(ptr) (IFF_ID((ptr)[4], (ptr)[5], (ptr)[6], (ptr)[7]))
#define IFF_chunk_contents(x) ((x) + 2*sizeof(uint_32))
#define IFF_chunk_end(x) (IFF_chunk_contents(x) + ROUND_UP (IFF_chunk_len (x)))

/* Dealing with inSANE data formats ... */

static double
IFF_read_extended (iff_handle handle)
{
  union
  {
    double res;
    char buf[8];
  } trick;

  int last_lo_threebits, i;
  
  /* Get sign and exponent */

  int wk0 = IFF_get_short (handle);

  /* Recode for double... */

  int sign = wk0 & 0x8000;
  int expt = wk0 & 0x7fff;

  expt -= 0x4000;		/* Take off excess */
  expt <<= 4;			/* Shift to reduce size */
  expt += 0x4000;		/* Put back excess */
  expt &= 0x7ff0;
  
  /* First byte of double is entirely from that ---
   * second byte has a bit of mantissa in it.
   */
  
  trick.buf[0] = expt >> 8;

  /* Second byte has some of the mantissa in it --- mask off hidden bit,
   * which *isn't* hidden in the extended representation.
   */
  
  wk0 = (*handle++) & 0x7f;
  trick.buf[1] = (expt & 0xf0) | (wk0 >> 3);

  /* Handle the rest of the mantissa */
  
  last_lo_threebits = wk0 & 0x07;

  for (i = 2; i < 8; ++i)
  {
    wk0 = *handle++;
    trick.buf[i] = (last_lo_threebits << 5) | (wk0 >> 5);
    last_lo_threebits = wk0 & 0x07;
  }

  /* Done */

  return trick.res;
}

char *iff_do_form(iff_handle, struct format *, iff_handle);
char *aiff_do_comm(iff_handle, struct format *, iff_handle);
char *aiff_do_ssnd(iff_handle, struct format *, iff_handle);

struct iff_handler_info
{
  uint_32 type;
  char *(*handler)(iff_handle, struct format *, iff_handle);
}
iff_handlers[] =
{
{ IFF_Form, iff_do_form },
{ AIFF_COMM, aiff_do_comm },
{ AIFF_SSND, aiff_do_ssnd },
};

#define IFF_NHANDLERS (sizeof(iff_handlers)/sizeof(struct iff_handler_info))

char *
iff_do_form (iff_handle handle, struct format *samp, iff_handle start)
{
  uint_32 type;
  iff_handle form_end;
  iff_handle form_start = handle;
  char dummy[] = "xxxx";
  char *errmsg = malloc(100);

  if (errmsg == NULL) return "Couldn't malloc\n";
  
  /* Initialize a few fields with insane values so we can detect
   * missing chunks...
   */
  
  samp->data_offset = 0;
  samp->nbits = 0;

  /* Verify the IFF Form header itself... */
  
  if (IFF_chunk_type (handle) != IFF_Form)
  {
    strncpy (dummy, &handle, 4);
    sprintf (errmsg, "Expecting FORM, got %s", dummy);
    return errmsg;
  }
  
  form_end = IFF_chunk_end (handle);
  handle = IFF_chunk_contents (handle);
  type = IFF_get_long (handle);
  
  if (type != IFF_AIFF && type != IFF_AIFC)
  {
    strncpy (dummy, &type, 4);
    sprintf (errmsg, "IFF-type file of unknown type %s", dummy);
    return errmsg;
  }

  free (errmsg);
  
  while (handle < form_end && handle < (start + HDR_LENGTH))
  {
    int i;
    
    iff_handle chunk_end = IFF_chunk_end (handle);
    uint_32 chunk_type = IFF_chunk_type (handle);

    for (i = 0; i < IFF_NHANDLERS; ++i)
      if (chunk_type == iff_handlers[i].type)
	break;

    if (i < IFF_NHANDLERS) iff_handlers[i].handler (handle, samp, start);

    handle = chunk_end;
  }

  if (samp->nbits == 0) return "Bad AIFF file --- missing COMM chunk!";
  if (samp->data_offset == 0)
  {
    if (type == IFF_AIFC) return "Can't handle compressed AIFF data";
    else return "Bad AIFF file --- can't find the sound!";
  }

  return NULL;
}

char *
aiff_do_comm (iff_handle handle, struct format *samp, iff_handle start)
{
  iff_handle end = handle + IFF_chunk_len (handle);

  handle = IFF_chunk_contents (handle);

  samp->channels = IFF_get_short (handle);
  samp->num_samples = IFF_get_long (handle);
  samp->nbits = IFF_get_short (handle);
  samp->rate = IFF_get_ext (handle);

  /* Round nbits up to even multiple of eight */
  
  samp->nbits = (samp->nbits + 7) & ~7;

  /* AIFF files only support linear encoding */

  samp->coding = CODE_LINEAR;
  samp->little_endian = 0;
  
  return NULL;
}

char *
aiff_do_ssnd (iff_handle handle, struct format *samp, iff_handle start)
{
  iff_handle chandle = IFF_chunk_contents (handle);
  int blocksize = IFF_get_long (chandle);
  int offset = IFF_get_long (chandle);

  /* Note that we set file_size to the end of the samples --- don't
   * want to wind up trying to play the contents of any ID chunks that
   * may follow the SSND.
   */
  
  samp->data_offset = (chandle + offset) - start;
  samp->file_size = (handle - start) + IFF_chunk_len (handle);
  
  return NULL;
}

char *
grok_aiff (struct format *samp, char *header)
{
  return iff_do_form (header, samp, header);
}
