#include <stdio.h>
#include <memory.h>
#include <malloc.h>

#include "audio.h"

/*
 * xplay --- X sound player.
 *
 * This program is copyright 1994 by Robert S. Thau.  It is offered as
 * is, and no warranty is offered of suitability or fitness for any
 * particular purpose; use at your own risk.  Contents may have settled
 * during shipment.  The program may be redistributed for noncommercial
 * purposes only, so long as all files contain a prominent notice of when
 * they were last changed, and who changed them.
 */

/* RIFF/WAVE --- like AIFF, only different (in particular, if
 * you look at the #defines, endianness is reversed).  Praise be
 * to Microsoft!!
 *
 * This won't work yet --- need to get endianness bit into the
 * format structure, and to make convert.c respect it.  Likewise
 * for bias, to handle the oddity that 8-bit PCM samples are *not*
 * zero-centered.
 */

typedef unsigned long uint_32;
#define ROUND_UP(x) ((((int)x)&1) ? (x) + 1 : (x))

/* These macros side-effect the args --- beware!
 *
 * Note also that they don't take advantage of the fact that the Sparc
 * has the right byte-sex, because the data may very well have the
 * wrong alignment.  If you've gotta go byte-by-byte anyway...
 *
 * But the function which reads "extendeds" *does* know the order
 * of bytes in a double.
 *
 * Note also that you don't get all the bytes of a value which is
 * stored in the file as an 'ext'.
 */

typedef unsigned char *iff_handle;

#define IFF_ID(a,b,c,d) (((d)<<24) + ((c)<<16) + ((b)<<8) + (a))

#define IFF_get_char(ptr) (*(ptr)++)
#define IFF_get_short(ptr) ((ptr) += 2, ((ptr)[-1] << 8) + (ptr)[-2])
#define IFF_get_long(ptr) ((ptr) += 4, \
			   IFF_ID((ptr)[-4], (ptr)[-3], (ptr)[-2], (ptr)[-1]))

static double IFF_read_extended (iff_handle);
				  
/* Chunk types we know about ... */

#define IFF_Form IFF_ID('R', 'I', 'F', 'F')
#define IFF_WAVE IFF_ID('W', 'A', 'V', 'E')

#define WAVE_FMT  IFF_ID('f', 'm', 't', ' ')
#define WAVE_DATA IFF_ID('d', 'a', 't', 'a')

#define WAVE_FORMAT_PCM 1
#define IBM_FORMAT_MULAW 0x0101
#define IBM_FORMAT_ALAW 0x0102

/* Handling chunks ... */

#define IFF_chunk_type(ptr) (IFF_ID((ptr)[0], (ptr)[1], (ptr)[2], (ptr)[3]))
#define IFF_chunk_len(ptr) (IFF_ID((ptr)[4], (ptr)[5], (ptr)[6], (ptr)[7]))
#define IFF_chunk_contents(x) ((x) + 2*sizeof(uint_32))
#define IFF_chunk_end(x) (IFF_chunk_contents(x) + ROUND_UP (IFF_chunk_len (x)))

char *riff_do_form(iff_handle, struct format *, iff_handle);
char *wave_do_fmt(iff_handle, struct format *, iff_handle);
char *wave_do_data(iff_handle, struct format *, iff_handle);

struct riff_handler_info
{
  uint_32 type;
  char *(*handler)(iff_handle, struct format *, iff_handle);
}
riff_handlers[] =
{
{ IFF_Form, riff_do_form },
{ WAVE_FMT,  wave_do_fmt },
{ WAVE_DATA, wave_do_data },
};

#define RIFF_NHANDLERS (sizeof(riff_handlers)/sizeof(struct riff_handler_info))

char *
riff_do_form (iff_handle handle, struct format *samp, iff_handle start)
{
  uint_32 type;
  iff_handle form_end;
  iff_handle form_start = handle;
  char dummy[] = "xxxx";
  char *errmsg = malloc(100);

  if (errmsg == NULL) return "Couldn't malloc\n";
  
  /* Initialize a few fields with insane values so we can detect
   * missing chunks...
   */
  
  samp->data_offset = 0;
  samp->nbits = 0;

  /* Verify the IFF Form header itself... */
  
  if (IFF_chunk_type (handle) != IFF_Form)
  {
    strncpy (dummy, &handle, 4);
    sprintf (errmsg, "Expecting RIFF, got %s", dummy);
    return errmsg;
  }
  
  form_end = IFF_chunk_end (handle);
  handle = IFF_chunk_contents (handle);
  type = IFF_get_long (handle);
  
  if (type != IFF_WAVE)
  {
    strncpy (dummy, &type, 4);
    sprintf (errmsg, "RIFF-type file of unknown type %s", dummy);
    return errmsg;
  }

  free (errmsg);
  
  while (handle < form_end && handle < (start + HDR_LENGTH))
  {
    int i;
    
    iff_handle chunk_end = IFF_chunk_end (handle);
    uint_32 chunk_type = IFF_chunk_type (handle);

    for (i = 0; i < RIFF_NHANDLERS; ++i)
      if (chunk_type == riff_handlers[i].type)
	break;

    if (i < RIFF_NHANDLERS) riff_handlers[i].handler (handle, samp, start);

    handle = chunk_end;
  }

  if (samp->nbits == 0) return "WAVE file not understood --- compressed???";
  if (samp->data_offset == 0)
  {
    return "Weird WAVE file --- can't find the sound!";
  }

  samp->little_endian = 1;
  
  free(errmsg);
  return NULL;
}

/* No code for the endianness or off-centering yet */

char *
wave_do_fmt (iff_handle handle, struct format *samp, iff_handle start)
{
  iff_handle end = handle + IFF_chunk_len (handle);
  int is_pcm = 0;
  long dummy;

  handle = IFF_chunk_contents (handle);

  switch (IFF_get_short (handle))
  {
  case WAVE_FORMAT_PCM: ++is_pcm; break;
    
  case IBM_FORMAT_MULAW: 
    samp->nbits = 8;
    samp->coding = CODE_ULAW;
    break;

  case IBM_FORMAT_ALAW:
    samp->nbits = 8;
    samp->coding = CODE_ALAW;
    break;
    
  default:			/* Leave nbits at nonsense values... */
    break;
  }
  
  samp->channels = IFF_get_short (handle);
  samp->rate = IFF_get_long (handle);
  dummy = IFF_get_long (handle); /* Avg bytes per sec */
  dummy = IFF_get_short (handle); /* Block align */
  
  if (is_pcm)
  {
    samp->nbits = IFF_get_short (handle);

    /* Round nbits up to even multiple of eight */
  
    samp->nbits = (samp->nbits + 7) & ~7;
    
    if (samp->nbits > 8)
      samp->coding = CODE_LINEAR;
    else
      samp->coding = CODE_LINEAR_UNSIGNED;
  }
  
  return NULL;
}

char *
wave_do_data (iff_handle handle, struct format *samp, iff_handle start)
{
  iff_handle chandle = IFF_chunk_contents (handle);
  int samp_size = (samp->channels * (samp->nbits / 8));

  if (samp_size == 0) samp_size = 1; /* Flagged as error elsewhere;
				      * must just avoid divide by zero.
				      */

  /* Note that we set file_size to the end of the samples --- don't
   * want to wind up trying to play the contents of any ID chunks that
   * may follow the data.
   */
  
  samp->data_offset = chandle - start;
  samp->file_size = samp->data_offset + IFF_chunk_len(handle);
  samp->num_samples =
    (samp->file_size - samp->data_offset)
      / samp_size;

  return NULL;
}

char *
grok_wave (struct format *samp, char *header)
{
  return riff_do_form (header, samp, header);
}
