#include <stdio.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/PanedW.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/Separator.h>
#include <Xm/LabelG.h>
#include <Xm/ArrowBG.h>
#include <Xm/ToggleB.h>

#include "play.h"

/*
 * xplay --- X sound player.
 *
 * This program is copyright 1994 by Robert S. Thau.  It is offered as
 * is, and no warranty is offered of suitability or fitness for any
 * particular purpose; use at your own risk.  Contents may have settled
 * during shipment.  The program may be redistributed for noncommercial
 * purposes only, so long as all files contain a prominent notice of when
 * they were last changed, and who changed them.
 */

void exit();

/* Resource foo: */
     
String fallback_resources[] =
{
  "*background: #B2B2B2",	/* Same as Netscape, to deal with
				 * its hogging of the color map...
				 */
  "*fontList: -*-helvetica-bold-r-*-*-*-120-*-*-*-*-*-*",
  "*XmToggleButton*indicatorSize: 14",
  "*XmToggleButton*selectColor: yellow",
  "*buttons*marginHeight:     0",
  "*buttons*spacing:  1",
  NULL
};

struct app_resources {
  int timeout;
  Bool detach;
  int seek_speed;
  char *port;
} app_resources;

#define res_offset(x) XtOffsetOf(struct app_resources, x)

/* Application resources.  Note that port defaults to "default", which
 * is understood to implement the actual default behavior in audio_start
 * in play.c --- this is so nothing hardware-specific drifts into this file.
 */

static XtResource resources[] =
{
{ "timeout", "Timeout", XtRInt, sizeof(int), res_offset(timeout),
    XtRImmediate, (XtPointer)3 },
{ "detach", "Detach", XtRBool, sizeof(Bool), res_offset(detach),
    XtRImmediate, (XtPointer)False },
{ "seekSpeed", "SeekSpeed", XtRInt, sizeof(int), res_offset(seek_speed),
    XtRImmediate, (XtPointer) 20 },
{ "port", "Port", XtRString, sizeof(String), res_offset(port),
    XtRImmediate, "default" },
};

static XrmOptionDescRec options[] =
{
{ "-timeout", "*timeout", XrmoptionSepArg, NULL },
{ "-forkoff", "*detach",  XrmoptionNoArg,  "True" },
{ "-port", "*port", XrmoptionSepArg, NULL},
{ "-stay", "*Stay*set", XrmoptionNoArg, "True" },
};

XtAppContext app;

/* Globals pertaining to graphic state --- bad style.
 *
 * Note that the housekeeping to maintain these is a little disorganized;
 * in general, I've just shoved the updates into the code that handles
 * changes to the display for the same condition.  So, for instance, once
 * we have the whole file, have_whole_file should be set, we should cancel
 * the Xt Input for the audio file to avoid burning cycles, and we *also*
 * want to stop displaying how much we've read.  The same code does all of
 * this (it's in update_time_label() below).
 */

int playing = 1;
int have_whole_file = 0;

int input_length;		/* Total length of the sample, as of
				 * last report from play.c.  Note that this
				 * can change, if play.c is reading a file
				 * with unknown size...
				 */
int input_id;			/* The Xt input ID for the audio, so that
				 * we can cancel it when we're finished.
				 */

Widget play_button;
Widget save_button;
Widget stay_button;
Widget quit_button;
Widget rw_button;
Widget ff_button;
Widget time_line;
Widget volume_slider;

/*****************************************************************
 *
 * Keeping track of dialog boxes...
 * don't want to die on a timeout while a dialog box is up, so we
 * need to keep track of how many are up.  Basically, everything
 * that creates a dialog box is supposed to call track_dialog,
 * which installs a bunch of callbacks on the dialog shell which
 * keep track of what's up (if anything).
 *
 * Isn't this *elegant*?  Hmmmph.
 */

int dialog_boxes_up = 0;

void note_dialog_popup (Widget w, void *dum, void *dum2) {++dialog_boxes_up;}
void note_dialog_popdown (Widget w, void *dum, void *dum2) {--dialog_boxes_up;}

void track_dialog (Widget dialog)
{
  Widget dialog_shell = XtParent(dialog);
  
  XtAddCallback (dialog_shell, XmNpopupCallback, note_dialog_popup, NULL);
  XtAddCallback (dialog_shell, XmNpopdownCallback, note_dialog_popdown, NULL);
}    

/*****************************************************************
 *
 * The end-of-play timeout ...
 *
 * If we run off the end of the audio sample that we're playing, and
 * the user does absolutely nothing to it for ten seconds or so, then
 * we want it to go away (lest xplay processes pile up).
 *
 * The trick is doing this in a way that doesn't wind up tracking mud
 * through the rest of the code (like, for instance, installing calls
 * to kill_timeout function in every damn callback in sight --- an
 * approach which finally hit the breaking point when I faced the
 * potential of installing callbacks which had no other purpose than
 * to call kill_timeout).
 */

void
timeout_event_loop ()
{
  static int timing_out = 0;	/* Don't reenter! */
  XtIntervalId timeout;
  
  if (!timing_out && dialog_boxes_up == 0
      && !XmToggleButtonGetState(stay_button))
  {
    XEvent ev;
    
    timeout = XtAppAddTimeOut (app, 1000 * app_resources.timeout, exit, 0);
    timing_out = 1;

    /* OK, timeout started.  Now, keep processing events as normal.
     * If one of them is an attempt by the user to submit input,
     * abort the timeout.
     */
    
    do
    {
      XtAppNextEvent (app, &ev);
      XtDispatchEvent (&ev);
    } while (ev.type != KeyPress && ev.type != KeyRelease
	     && ev.type != ButtonPress && ev.type != ButtonRelease);

    XtRemoveTimeOut(timeout);
    timing_out = 0;
  }
}

/*****************************************************************
 * Callback for quit button.  Can't *just* use exit() here, as for
 * the timeout, as sound might still be playing.
 */

void
die (Widget dummy, void *dummy2, void*dummy3)
{
  audio_abort ();
  exit (0);
}

/*****************************************************************
 *
 * Callbacks and support for the time-line widget
 */

void update_time_label();

static int label_update_ct = 0;     
     
void
update_time_line ()
{
  int posn = audio_posn();

  if (posn > input_length - 10 || ++ label_update_ct > 9)
  {
    input_length = audio_size();
    if (posn > input_length) input_length = posn; /* Can't happen */
    
    XtVaSetValues (time_line, XmNmaximum, input_length, NULL);

    label_update_ct = 0;
    update_time_label();
  }
    
  if (posn >= 0)		/* Make sure that we are playing ... */
  {
    XmScaleSetValue (time_line, posn);

    if (audio_is_over ())
    {
      audio_abort();
      timeout_event_loop();
    }
  }
  
  XtAppAddTimeOut (app, 500, update_time_line, NULL);
}

/*
 * Mute sound playback while the user drags the slider...
 */

void
note_timeline_drag (Widget scale, void *dummy, void *acbs)
{
  if (playing) audio_abort();
}

void
note_timeline_adjust (Widget scale, void *dummy, void *acbs)
{
  XmScaleCallbackStruct *cbs = acbs;
    
  if (playing) audio_start (cbs->value, app_resources.port);
}

/*****************************************************************
 * Progress reports.
 */

char*
ptime (int ntenths)
{
  int secs = (ntenths + 9) / 10;
  static char buffer[100];
  
  if (secs <= 90)
  {
    sprintf (buffer, "%d sec", secs);
  }
  else
  {
    int mins = secs / 60;

    secs = secs - mins * 60;

    if (mins < 10)
      sprintf (buffer, "%d min, %d sec", mins, secs);
    else
    {
      mins += (secs >= 30);
      sprintf (buffer, "%d min", mins);
    }
  }
  
  return buffer;
}
     

void
update_time_label ()
{
  char buffer[100];

  if (have_whole_file) return;	/* It's not changing */
  
  if (audio_size_is_a_guess())
    strcpy (buffer, "Unknown size");
  else
    strcpy (buffer, ptime (audio_size()));

  if (!audio_is_all_read())
  {
    strcat (buffer, "; ");
    strcat (buffer, ptime (audio_last_posn_read()));
    strcat (buffer, " read");
  }
  else
  {
    have_whole_file = 1;	/* Don't do this again. */
    XtRemoveInput (input_id);
  }
  
  XtVaSetValues (time_line, XtVaTypedArg,
		 XmNtitleString, XmRString, buffer, strlen(buffer), NULL);
}

/*****************************************************************
 * Callbacks and support for rewind and fast-forward.
 *
 * The workhorse here is do_seek_timeout, which is called ten times
 * a second while rw or ff are pushed.  The direct widget callbacks
 * are start_seeking and end_seeking; they set up or tear down the
 * timeout.
 */

XtIntervalId seek_timeout;
int seek_incr;

void do_seek_timeout ()
{
  int scale_val;
  
  XmScaleGetValue (time_line, &scale_val);
  scale_val += seek_incr;
  
  if (scale_val < 0) scale_val = 0;
  else if (scale_val > audio_last_posn_read())
    scale_val = audio_last_posn_read();
  
  XmScaleSetValue (time_line, scale_val);

  seek_timeout = XtAppAddTimeOut (app, 100, do_seek_timeout, NULL);
}

void start_seeking (Widget w, void *dummy, void *other_dummy)
{
  /* Mute while seeking, as with direct scale adjustments... */
    
  if (playing) audio_abort();	
  
  /* Scaling foo ---
   *
   * seek_speed is a multiplier --- we want to seek that much faster
   * than actual playback.
   *
   * seek_incr is the number of scale-ticks (in tenths of a second)
   * we need to move in a tenth of a second to implement that rate.
   *
   * Conveniently, these work out to the same number...
   */
  
  if (w == rw_button) seek_incr = -app_resources.seek_speed;
  else seek_incr = app_resources.seek_speed;

  do_seek_timeout();
}

void end_seeking (Widget w, void *dummy, void *other_dummy)
{
  int scale_val;

  XtRemoveTimeOut (seek_timeout);
  XmScaleGetValue (time_line, &scale_val);
  
  if (playing) audio_start (scale_val, app_resources.port);
}

/*****************************************************************
 * Callbacks and support for the pause/play button
 */

void
start_playing ()
{
  int when;
  
  XtVaSetValues (play_button, XtVaTypedArg,
		 XmNlabelString, XmRString, "Pause", 5, NULL);
  XFlush (XtDisplay (play_button));

  XmScaleGetValue (time_line, &when);
    
  audio_start (when, app_resources.port);
  
  XtAppAddTimeOut (app, 500, update_time_line, NULL);
  
  playing = 1;
}

void
stop_playing ()
{
  XtVaSetValues (play_button, XtVaTypedArg,
		 XmNlabelString, XmRString, "Play", 4, NULL);
  XFlush (XtDisplay (play_button));

  audio_abort ();
  playing = 0;
}

void
pause_or_play ()
{
  if (playing) stop_playing(); else start_playing();
}

/*****************************************************************
 *
 * Callbacks and support for the volume slider.
 */

void
note_volume_adjust (Widget scale, void *dummy, void *acbs)
{
  XmScaleCallbackStruct *cbs = acbs;
    
  audio_set_volume (cbs->value);
}

/*****************************************************************
 * 
 * Callbacks and support for "Save as ... "
 */

void error_from_parent (Widget parent, char *string)
{
  Widget dialog;
  Arg arg[1];
  XmString t;

  t = XmStringCreateSimple(string);
  XtSetArg (arg[0], XmNmessageString, t);
  dialog = XmCreateErrorDialog (parent, "oops", arg, 1);
  XmStringFree (t);

  track_dialog (dialog);
  
  /* No help for the helpless, and no difference between OK and Cancel ... */
  
  XtUnmanageChild (XmMessageBoxGetChild (dialog, XmDIALOG_HELP_BUTTON));
  XtUnmanageChild (XmMessageBoxGetChild (dialog, XmDIALOG_OK_BUTTON));

  XtManageChild (dialog);
}

void warning_from_parent (Widget parent, char *string,
			  XtCallbackProc callback)
{
  Widget dialog;
  Arg arg[1];
  XmString t;

  t = XmStringCreateSimple(string);
  XtSetArg (arg[0], XmNmessageString, t);
  dialog = XmCreateWarningDialog (parent, "oops", arg, 1);
  XmStringFree (t);

  track_dialog (dialog);

  /* No help for the helpless ... */
  
  XtUnmanageChild (XmMessageBoxGetChild (dialog, XmDIALOG_HELP_BUTTON));
  XtAddCallback (dialog, XmNokCallback, callback, NULL);

  XtManageChild (dialog);
}

void save_to_file (Widget filebox, void *dummy, void *acbs)
{
  char *filename;
  char *err_string;
  XmFileSelectionBoxCallbackStruct *cbs = acbs;

  if (!XmStringGetLtoR (cbs->value, XmSTRING_DEFAULT_CHARSET, &filename))
    return;

  if (!*filename) return;

  if (err_string = audio_save (filename))
    error_from_parent (save_button, err_string);

  free (filename);
}

void popup_file_box (Widget button, void *dummy1, void *dummy2)
{
  Arg arg[1];
  Widget dialog;

  XtSetArg (arg[0], XmNautoUnmanage, 1);
  
  dialog = XmCreateFileSelectionDialog (button, "filebox", arg, 1);
  
  track_dialog (dialog);

  XtUnmanageChild (XmFileSelectionBoxGetChild (dialog, XmDIALOG_HELP_BUTTON));
  
  XtAddCallback (dialog, XmNokCallback, save_to_file, NULL);

  XtManageChild (dialog);
}

void do_save_as (Widget button, void *dummy, void *dummy2)
{
  if (audio_is_all_read()) popup_file_box(button, NULL, NULL);
  else warning_from_parent (button, "Transfer not yet complete; save anyway?",
			    popup_file_box);
}

/****************************************************************
 *
 * The "Stay" toggle... really want a Motif 2.0 tri-state toggle,
 * with the third state indicating stay off and the timeout *on*;
 * that would give people a valuable clue as to what the button
 * does.
 *
 * In the meantime, we need this stuff to even make it behave
 * like other buttons...
 */

struct colors
{
  Pixel bg;
  Pixel topshadow;
  Pixel bottomshadow;
};

void switch_color_cb (Widget toggle, void *acolors, void *dummy)
{
  struct colors *colors = acolors;
  
  XtVaSetValues (toggle,
		 XmNbackground, colors->bg,
		 XmNtopShadowColor, colors->topshadow,
		 XmNbottomShadowColor, colors->bottomshadow,
		 NULL);
}

void setup_color_callbacks (Widget toggle)
{
  Pixel bg, topshadow, bottomshadow, select_color, fg;
  Colormap cmap;

  struct colors
    *disarmed = (struct colors *)XtMalloc (sizeof (struct colors)),
    *armed = (struct colors *)XtMalloc (sizeof(struct colors));

  XtVaGetValues (toggle, XmNbackground, &bg, XmNcolormap, &cmap, NULL);

  XmGetColors (XtScreen(toggle), cmap, bg,
	       &fg, &topshadow, &bottomshadow, &select_color);

  disarmed->bg = bg;
  disarmed->topshadow = topshadow;
  disarmed->bottomshadow = bottomshadow;

  armed->bg = select_color;
  armed->topshadow = bottomshadow;
  armed->bottomshadow = topshadow;

  XtAddCallback (toggle, XmNarmCallback, switch_color_cb, (void*)armed);
  XtAddCallback (toggle, XmNdisarmCallback, switch_color_cb, (void*)disarmed);
}

/****************************************************************
 *
 * Setup and misc...
 */

void
gui_init (Widget toplevel)
{
  /* It would be interesting to do a point-by-point comparison between
   * OS/360 JCL and this goop ...
   */
    
  Widget rowcol = XtVaCreateWidget ("rowcol", xmRowColumnWidgetClass, toplevel,
				  XmNorientation, XmHORIZONTAL,
				  XmNpacking, XmPACK_TIGHT,
				  XmNnumColumns, 1,
				  XmNentryAlignment, XmALIGNMENT_CENTER,
				  NULL);

  Widget paned = XtVaCreateWidget ("paned", xmPanedWindowWidgetClass, rowcol,
				   XmNsashWidth, 1, XmNsashHeight, 1, NULL);

  
#ifdef NOTDEF
  Widget vcol = XtVaCreateWidget ("vcol",  xmRowColumnWidgetClass, rowcol,
				  XmNorientation, XmVERTICAL,
				  XmNpacking, XmPACK_TIGHT,
				  XmNnumColumns, 1,
				  XmNentryAlignment, XmALIGNMENT_CENTER,
				  NULL);

  Widget vol_up = XtVaCreateManagedWidget
    ("volup", xmArrowButtonGadgetClass, vcol,
     XmNarrowDirection, XmARROW_UP, NULL);

  Widget vol_lab = XtVaCreateManagedWidget
    ("Vol", xmLabelGadgetClass, vcol, NULL);
  
  Widget vol_dn = XtVaCreateManagedWidget
    ("voldn", xmArrowButtonGadgetClass, vcol,
     XmNarrowDirection, XmARROW_DOWN, NULL);
#endif
  
  Widget vscal = XtVaCreateWidget ("volume", xmScaleWidgetClass, rowcol,
				   XtVaTypedArg, XmNtitleString, XmRString,
				     "V\nO\nL", 5,
				   XmNorientation, XmVERTICAL,
				   XmNminimum, audio_min_volume(),
				   XmNmaximum, audio_max_volume(),
				   XmNheight, 40,
				   NULL);
  
  Widget scale = XtVaCreateWidget ("scale", xmScaleWidgetClass, paned,
				   XtVaTypedArg, XmNtitleString, XmRString,
				     "Adjust slider to rewind", 23,
				   XmNorientation, XmHORIZONTAL,
				   XmNminimum, 0, XmNmaximum, input_length,
				   NULL);

  Widget buttons = XtVaCreateWidget ("buttons", xmRowColumnWidgetClass, paned,
				  XmNorientation, XmHORIZONTAL,
				  XmNpacking, XmPACK_TIGHT,
				  XmNnumColumns, 1,
				  XmNentryAlignment, XmALIGNMENT_CENTER,
				  NULL);

  rw_button = XtVaCreateManagedWidget ("rewind",
				  xmPushButtonWidgetClass, buttons,
				  XtVaTypedArg, XmNlabelString, XmRString,
				       "<<", 2, NULL);
  ff_button = XtVaCreateManagedWidget ("fast_forward",
				  xmPushButtonWidgetClass, buttons,
				  XtVaTypedArg, XmNlabelString, XmRString,
				       ">>", 2, NULL);

  /* Note the use of the barely-documented XmNrecomputeSize resource to
   * keep the Play/Pause button from changing its shape when its state
   * is toggled.  Without this, not only does play/pause change shape,
   * but the save and quit buttons shimmy to accomodate the change in
   * size.  UGLY.
   */
  
  play_button = XtVaCreateManagedWidget ("Pause",
				  xmPushButtonWidgetClass, buttons,
				  XmNrecomputeSize, False,
				  NULL);

  XtVaCreateManagedWidget (" ", xmLabelGadgetClass, buttons, NULL);
  
  save_button = XtVaCreateManagedWidget ("Save",
				  xmPushButtonWidgetClass, buttons, NULL);
  quit_button = XtVaCreateManagedWidget ("Quit",
				  xmPushButtonWidgetClass, buttons, NULL);
  stay_button = XtVaCreateManagedWidget ("Stay",
				  xmToggleButtonWidgetClass, buttons,
				  XmNshadowThickness, 2,
				  NULL);

  XtAddCallback (play_button, XmNactivateCallback, pause_or_play, NULL);
  XtAddCallback (quit_button, XmNactivateCallback, die, NULL);
  XtAddCallback (save_button, XmNactivateCallback, do_save_as, NULL);
  
  XtAddCallback (rw_button, XmNarmCallback, start_seeking, NULL);
  XtAddCallback (rw_button, XmNdisarmCallback, end_seeking, NULL);
  XtAddCallback (ff_button, XmNarmCallback, start_seeking, NULL);
  XtAddCallback (ff_button, XmNdisarmCallback, end_seeking, NULL);
  
  XtAddCallback (scale, XmNdragCallback, note_timeline_drag, NULL);
  XtAddCallback (scale, XmNvalueChangedCallback, note_timeline_adjust, NULL);
  
  XtAddCallback (vscal, XmNdragCallback, note_volume_adjust, NULL);
  XtAddCallback (vscal, XmNvalueChangedCallback, note_volume_adjust, NULL);
  
  setup_color_callbacks (stay_button);
  
  time_line = scale;
  volume_slider = vscal;
  
  XtManageChild (scale);
  XtManageChild (buttons);
  XtManageChild (paned);
  
#if 1
  XtManageChild (vscal);
#else
  XtManageChild (vcol);
#endif  
  
  XtManageChild (rowcol);
}

main (int argc, char **argv)
{
  char *errmsg = NULL, *filename = NULL, *progname = argv[0];
  char errbuffer[100];

  Widget toplevel = XtVaAppInitialize (&app, "XPlay",
				       options, XtNumber(options),
				       &argc, argv,
				       fallback_resources, NULL);
  
  XtVaGetApplicationResources (toplevel, &app_resources,
			       resources, XtNumber(resources), NULL);
  
  if (argc > 2)
  {
    fprintf (stderr, "Usage: %s [file]\n", progname);
    exit (1);
  }
  
  if (argc == 1) errmsg = "Xplay: No input specified";
  else
  {
    filename = argv[1];
  
    errmsg = audio_file_init (filename);
  
    if (errmsg == NULL && !audio_good_port(app_resources.port))
    {
      sprintf (errbuffer, "No such audio output port: %s", app_resources.port);
      errmsg = errbuffer;
    }
  }
  
  /* Netscape bug compensation... */
  if (app_resources.detach && fork() > 0) exit(); 
  
  if (errmsg)
  {
    Widget m = XtVaCreateManagedWidget ("Oops",
				 xmMessageBoxWidgetClass, toplevel,
				 XmNdialogType, XmDIALOG_ERROR,
				 XtVaTypedArg, XmNmessageString, XmRString,
				     errmsg, strlen(errmsg),
				 NULL);
				 
    XtUnmanageChild (XmMessageBoxGetChild (m, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild (XmMessageBoxGetChild (m, XmDIALOG_OK_BUTTON));
    
    XtAddCallback (m, XmNcancelCallback, exit, (void*)1);
  }
  else
  {
    /* The only purpose XtInputReadMask has is to be passed as a magic
     * cookie to XtAppAddInput --- but naturally, you've got to cast it.
     * Look at the source code for XtAppAddInput itself, and you find
     * that it immediately casts it back to the integer it always was.
     * What idiot is responsible for this?
     */
    
    input_id =  XtAppAddInput (app, audio_fd(), (XtPointer)XtInputReadMask,
			       audio_read_some_more, NULL);
    input_length = audio_size();	/* Beware!  This can change! */
    gui_init (toplevel);
    start_playing();

    XmScaleSetValue (volume_slider, audio_current_volume());
    update_time_label();
  }

  XtRealizeWidget (toplevel);
  XtAppMainLoop (app);
}
