#include <stdio.h>
#include <malloc.h>

/*
 * xplay --- X sound player.
 *
 * This program is copyright 1994 by Robert S. Thau.  It is offered as
 * is, and no warranty is offered of suitability or fitness for any
 * particular purpose; use at your own risk.  Contents may have settled
 * during shipment.  The program may be redistributed for noncommercial
 * purposes only, so long as all files contain a prominent notice of when
 * they were last changed, and who changed them.
 */

/*
 * Welcome to the dark heart of Xblare.
 *
 * This is the central audio player code, including the bits that deal
 * directly with the kernel audio driver.  These *call* the
 * file-format handlers and conversion routines ... the reason is that
 * we want to do the conversion on demand, as the audio device needs
 * data; therefore, the conversion has to be driven by the routines
 * which actually talk to the device.
 */

#include <signal.h>
#include <fcntl.h>
#include <stropts.h>
#include <errno.h>

#include "file.h"
#include "audio.h"
#include "convert.h"
#include "play.h"

/* #define STDOUT_DUMP */

/* Globals to this module --- sound format structures representing the
 * format of the incoming sound stream before and after conversion;
 * also, the xfile containing the data itself, and the audio device
 * file handle (if any).
 */

static struct xfile src_xfile;
static struct format src_format;
static struct format dst_format;
void *conversion_cookie;

static int audio_device = -1;

/* Handling underruns --- the audio playing code is driven by SIGIO's
 * from the kernel audio driver telling us it wants more data.  If we
 * don't have anything to give it, this is a problem, since play.c can't
 * hang (if it does, X events go unhandled, which is bad all around).
 *
 * So, if it looks like we're in such a pickle, what we do is to
 * set the underrun flag and stop playing sound; we then start checking
 * to see if more data has arrived and we can restart the audio transfer
 * loop (feed_device) below.
 *
 * This means that something in here has to do the underrun check
 * periodically.  Since the GUI is supposed to be periodically calling
 * audio_read_some_more() as long as there is data left to be read (i.e.,
 * so long as underrun conditions are possible), we put the check there.
 * It's a kludge, but it works.
 *
 * Here are declarations of the flag we check, and the code to invoke if
 * it's high...
 */

int underrun = 0;
static void attempt_underrun_recovery();
   
/*****************************************************************
 *
 * File handling.  For reasons explained above, we want all handling
 * of the file to go through this layer, so these functions are here
 * for completeness...
 */

char *audio_file_init (char *filename)
{
  char *errmsg;
  int file_size;
  
  if (errmsg = open_xfile (&src_xfile, filename)) return errmsg;

  while (src_xfile.total_bytes_read < HDR_LENGTH && !src_xfile.eof)
    read_whats_available (&src_xfile);
  
  if (src_xfile.total_bytes_read == 0) return "Input file was empty";
  
  if (errmsg = grok_audio_header (&src_format, src_xfile.first->data))
    return errmsg;
  
  if (errmsg = check_for_conversion (&src_format))
    return errmsg;

  return NULL;
}

char *audio_save (char *filename)
{
  return save_xfile (&src_xfile, filename);
}

/*****************************************************************
 * A little trivial accounting.  These are only defined after
 * audio_file_init().  audio_posn() (below) is not supposed to
 * exceed audio_size(), but *beware* if audio_size() itself is
 * changing...
 */

static int bogus_size_estimate = 50;

int audio_last_posn_read()
{
  return (samples_from_size (&src_format,
			     src_xfile.total_bytes_read) * 10
	  / src_format.rate);
}

int audio_size ()
{
  if (src_format.num_samples > 0) 
    return (src_format.num_samples * 10) / src_format.rate;
    
  return audio_last_posn_read(); /* OK, we're desperate */
}

int audio_size_is_a_guess()
{
  return (src_format.num_samples <= 0);
}

int audio_is_all_read ()
{
  return src_xfile.eof;
}

int audio_fd ()
{
  return src_xfile.fd;
}

void audio_read_some_more ()
{
  read_whats_available (&src_xfile);

  if (src_xfile.eof)
  {
    /* If we didn't know file size or duration before, we do now */
       
    if (src_format.file_size < 0)
      src_format.file_size = src_xfile.total_bytes_read;

    if (src_format.num_samples < 0)
      src_format.num_samples = samples_from_size (&src_format,
						  src_xfile.total_bytes_read);
  }

  if (underrun) attempt_underrun_recovery();
}

/****************************************************************
 *
 * Actual Sun-dependant stuff begins here.
 *
 * The only significant code below this line which is NOT dependant
 * on the details of interfacing to the Sun audio device driver is
 * the buffer-handling code, including most of the transfer loop.
 */

#include <sys/ioctl.h>

#ifdef SVR4
#include <sys/audioio.h>
#else
#include <sun/audioio.h>
#endif

/*****************************************************************
 *
 * Descriptions of the actual audio hardware...
 * This stuff is here so that convert.c doesn't have to know anything
 * about the hardware.  
 *
 * The devspec arrays are general enough to specify the Sun sound chips,
 * but not more (not yet, anyway).
 *
 * arr[0] --- encoding the chip supports.  If ULAW, it is understood to
 *            imply native support of ALAW as well.  LINEAR is understood
 *            to be sixteen bit linear, and to imply a superset of ULAW.
 * arr[1] --- maximum number of channels.  Either one or two.
 * arr[2] --- number of bits... assumed the device can support any less.
 * arr[2..] --- data rates supported, in Hz, sorted & terminated by zero.
 */

int devspec_amd[] = { CODE_ULAW, 1, 8, 8000, 0 };
int devspec_dbri[] = { CODE_LINEAR, 2, 16,
			 8000, 9600, 11025, 16000, 18900, 22050,
			 32000, 37800, 44100, 48000, 0 };

/* The device on Sparc 5s identifies itself as yet another weird thing.
 * I have no idea what it is, but I do know that it supports 48 Khz
 * 16-bit linear.  So, we convert everything to that --- burns a little
 * extra CPU, but it's known to work...
 */

int devspec_sparc5[] = { CODE_LINEAR, 2, 16, 8000, 48000, 0 };

/* Output ports... */

struct port_table
{
  char *name; int mask;
}
port_table[] =
{
{ "default", 0 },		/* Makes audio_start use whatever we got */
{ "speaker", AUDIO_SPEAKER },
{ "headphone", AUDIO_HEADPHONE },
{ "line_out", AUDIO_LINE_OUT },
};

int port_mask (char *port_name)
{
  int i;
  
  for (i = 0; i < sizeof(port_table)/sizeof(port_table[0]); ++i)
    if (!strcmp (port_name, port_table[i].name))
      return port_table[i].mask;

  return -1;
}

int audio_good_port (char *port_name)
{
  return port_mask (port_name) >= 0;
}

/*****************************************************************
 *
 * Accounting for our position in the audio stream... the kernel audio
 * device driver keeps track of two counts:
 *
 * 1) the total number of samples played
 * 2) the total number of zero-length 'EOF' packets played.
 *
 * The latter mechanism is intended as a quick and easy way of
 * determining when the speaker has actually reached a certain
 * point --- you write a zero-length 'EOF' packet into the data
 * stream, and when it actually gets written to the user, the
 * eof count in the audio info structure goes up.  We write such
 * EOF packets when we reach the end of the sample (see feed_device).
 *
 * The variables audio_last_start and audio_last_start_eof store
 * biases on these counts.  'ainfo' contains the current values of
 * the counts, and is updated on each call to audio_posn().
 *
 * Note that the biases go in opposite ways because AUDIO_SETINFO
 * can't seem to reset the EOF count.
 */

static int audio_last_start = 0;
static int audio_last_start_eof = 0;

static audio_info_t ainfo;

int
audio_posn ()
{
  int dst_sample_posn;
  int val;
  
  if (audio_device < 0) return -1;
  
  ioctl (audio_device, AUDIO_GETINFO, &ainfo);

  dst_sample_posn = audio_last_start + ainfo.play.samples;
  val = (dst_sample_posn * 10) / dst_format.rate;

  return (val > 0) ? val : 0;
}

/* Cute glitch --- this is only defined after a successful call to audio_posn
 */

int audio_is_over ()
{
  return (ainfo.play.eof > audio_last_start_eof);
}

/* This rountine is called by audio_start() to ensure that sample
 * counts and EOF information reported by the above routines are
 * correct.  Note that it must be called with the audio device stopped
 * (i.e., internal kernel queues flushed).
 */

void audio_reset_counts (int offset)
{
  offset = offset * dst_format.rate / 10;
  
  ioctl (audio_device, AUDIO_GETINFO, &ainfo);
  
  audio_last_start = offset;
  audio_last_start_eof = ainfo.play.eof;

  ainfo.play.samples = 0;
  ioctl (audio_device, AUDIO_SETINFO, &ainfo);
}

void adj_sample_count (int offset)
{
  audio_last_start += offset;
}

/*****************************************************************
 *
 * The volume stuff.  Note that we have to deal with the user frobbing
 * the volume slider when we are paused, so we maintain a shadow variable.
 *
 * audio_start() calls audio_current_volume() to get a value for the gain
 * field of its audio_info structure; the first time through, we grab the
 * value already there to set this all up.
 */

int remembered_volume = -1;

int audio_min_volume() { return AUDIO_MIN_GAIN; }
int audio_max_volume() { return AUDIO_MAX_GAIN; }

int audio_current_volume()
{
  if (remembered_volume < 0)
  {
    audio_info_t ainfo;
    ioctl (audio_device, AUDIO_GETINFO, &ainfo);
    remembered_volume = ainfo.play.gain;
  }

  return remembered_volume;
}

void audio_set_volume(int vol)
{
  remembered_volume = vol;

  if (audio_device >= 0)
  {
    audio_info_t ainfo;
    
    AUDIO_INITINFO(&ainfo);
    ainfo.play.gain = vol;
    ioctl (audio_device, AUDIO_SETINFO, &ainfo);
  }
}

/*****************************************************************
 * Buffering...
 *
 * We double-buffer the converted samples with about a second's worth
 * of data per buffer.  At any given time, while playing, we are
 * simultaneously playing data out of current_buffer, while converting
 * samples from the input stream into next_buffer.  When current_buffer
 * is done, the two switch places, unless next_buffer isn't finished yet;
 * if that ever happens, we do the underrun dance (see above...).
 */

#define AUDIO_BUF_SIZE 8192

struct buffer
{
  int length;			/* May be < AUDIO_BUF_SIZE if close to EOF */
  int buf_offset;		/* Index of first byte *not* sent to device */
  char *writp;			/* Index of first byte not yet read */
  int dst_offset;		/* Index of the bytes in here relative
				 * to entire destination stream.
				 */
    
  char data [AUDIO_BUF_SIZE];
};

static struct buffer buf0;
static struct buffer buf1;

static struct buffer *current_buffer;
static struct buffer *next_buffer;

static struct xfile_chunk *xp;
static char *srcp;		/* Pointer to first xfile byte not read. */

static void
fill_buffer (struct buffer *buf)
{
  int dst_len = buf->length - (buf->writp - buf->data);
  int src_len;

  src_len = xp->good_data_end - srcp;
  
  if (src_len > 0 && dst_len > 0)
  {
    convert_samples (&buf->writp, &srcp,
		     conversion_cookie,
		     dst_len, src_len);
  }
  
  if (src_format.file_size > 0 &&
      (src_format.file_size - xp->offset_in_file - (srcp - xp->data)) < 10)
  {
    /* EOF, or close enough that we don't care
     */
  
    buf->length = buf->writp - buf->data;
  }
  else if (srcp == xp->data_end && xp->next)
  {
    /* End of buffer, but not EOF */

    xp = xp->next;
    srcp = xp->data;
    fill_buffer(buf);
  }
}

static int
buffer_done (struct buffer *buf)
{
  return buf->length == (buf->writp - buf->data);
}

static void
start_buffer (struct buffer *buf, int dst_offset)
{
  buf->dst_offset = dst_offset;
  buf->length = AUDIO_BUF_SIZE; /* For now... gets reset if we see EOF */
  buf->buf_offset = 0;
  buf->writp = buf->data;

  fill_buffer (buf);
}

static void
init_buffers (int offset)
{
  int my_off = offset_for_sample(&src_format, offset);

  for (xp = src_xfile.first;
       (xp->next != NULL) && (my_off > xp->data_end - xp->data);
       xp = xp->next)
    my_off -= xp->data_end - xp->data;

  srcp = xp->data + my_off;
  
  current_buffer = &buf0;
  next_buffer = &buf1;

  /* We may not have enough data to fill a buffer at this point,
   * so we don't chance it --- just simulate an underrun condition,
   * and all will be sane when we "recover".
   */
  
  underrun = 1;
  start_buffer (next_buffer, offset);
  attempt_underrun_recovery();
}

/*****************************************************************
 *
 * The audio transfer loop...
 *
 * The Sun audio device is able to send us a signal when it wants
 * more data.  We use this in conjunction with the double-buffering
 * above to manage transfers into the device without having to use
 * X timeouts, idle procedures, or similar gimmickry.
 *
 * feed_device plays the role of the lower half of a hardware device
 * driver, shoving as much as it can into the kernel audio queue, and
 * arranging for an interrupt (SIGPOLL) when it wants more.
 *
 * audio_start and audio_abort play the role of the upper half,
 * setting up non-volatile device control registers and getting the
 * interrupt-driven transfer loop going (and then controlling it).
 */


static void
feed_device ()
{
  int len_written;

  len_written = 
    write (audio_device, current_buffer->data + current_buffer->buf_offset,
	   current_buffer->length - current_buffer->buf_offset);

#ifdef STDOUT_DUMP
  write (1, current_buffer->data + current_buffer->buf_offset, len_written);
#endif

  /* Note that len_written can be -1 if there was no space
   * in the audio device's STREAMS buffers --- far more likely
   * below than here, but we check both places just in to be safe.
   */

  if (len_written > 0) current_buffer->buf_offset += len_written;

  fill_buffer (next_buffer);
  
  /* Check if we're done with this buffer... */

  while (current_buffer->buf_offset == current_buffer->length)
  {
    struct buffer *temp;
    int retries;

    /* Yep.  Try to move to next_buffer, finishing up if need be */

    for (retries = 0; !buffer_done (next_buffer) && retries < 10; ++retries)
      fill_buffer (next_buffer);
    
    if (!buffer_done (next_buffer))
    {
      /* No data!  Do the underrun dance... */
      
      underrun = 1;
      return;
    }
    
    temp = current_buffer;
    current_buffer = next_buffer;
    next_buffer = temp;
      
    /* Start on new next buffer */
    
    start_buffer (next_buffer,
		  current_buffer->dst_offset + current_buffer->length);
    
    /* Write as much of the new current buffer as we can fit ...
     */
    
    len_written =
      write (audio_device, current_buffer->data + current_buffer->buf_offset,
	     current_buffer->length - current_buffer->buf_offset);

#ifdef STDOUT_DUMP
    write (1, current_buffer->data + current_buffer->buf_offset, len_written);
#endif

    if (len_written > 0) current_buffer->buf_offset += len_written;
    
    /* If the new current buffer was empty, we've reached EOF
     * (and incidentally, we just wrote a zero-length EOF packet).
     *
     * Shut down the transfer loop.
     */

    if (current_buffer->length == 0)
    {
      ioctl (audio_device, I_SETSIG, 0);
      break;
    }
  }
}

void
attempt_underrun_recovery()
{
  if (audio_device < 0)
  {
    underrun = 0; /* We aborted. */
    return;
  }

  fill_buffer (next_buffer);
  
  if (buffer_done(next_buffer))
  {
    struct buffer *temp = current_buffer;

    current_buffer = next_buffer;
    next_buffer = temp;
    
    start_buffer (next_buffer,
		  current_buffer->dst_offset + current_buffer->length);

    underrun = 0;
    feed_device();
  }
}

/*****************************************************************
 *
 * Controlling the transfer loop.
 */

void
audio_start (int offset, char *port)
{
  audio_info_t info;

#ifdef SVR4
  struct sigaction sig_act;
#endif

  int portm = port_mask (port);
    
  /* Sanity check on the offset... don't want to have the user waiting
   * half an hour for the net to reach the point where he'd flicked the
   * slider...
   */
  
  if (offset > audio_last_posn_read()) offset = audio_last_posn_read();
  
  /* Set up (or reset) audio device */
  
  if (audio_device < 0)
  {
    /* Need to open the audio device... */
    
    AUDIO_INITINFO(&info);
    
    audio_device = open ("/dev/audio", O_WRONLY|O_NONBLOCK);
    
    if (audio_device < 0)
    {
      perror("/dev/audio"); exit(1);
    }
    
    if (conversion_cookie == NULL)
    {
      int devtype;
      int *devspec = devspec_amd;

#ifdef SVR4
      audio_device_t dev_id;

      ioctl (audio_device, AUDIO_GETDEV, &dev_id);
      
      if (!strcmp (dev_id.name, "SUNW,dbri")) devspec = devspec_dbri;
      else if (!strcmp (dev_id.name, "SUNW,CS4231")) devspec = devspec_sparc5;
#else
      ioctl (audio_device, AUDIO_GETDEV, &devtype);
      
      if (devtype == AUDIO_DEV_SPEAKERBOX) devspec = devspec_dbri;
      else if (devtype == 5) devspec = devspec_sparc5;
#endif

      conversion_cookie = grok_for_conversion (&dst_format, &src_format,
					       devspec);

      if (!conversion_cookie)
      {
	perror ("Couldn't malloc!!!"); exit(1);
      }
    }

    info.play.sample_rate = dst_format.rate;
    info.play.channels = dst_format.channels;
    info.play.precision = dst_format.nbits;
    info.play.gain = audio_current_volume();
    
    info.play.encoding = dst_format.coding;

    switch (dst_format.coding)
    {
    case CODE_ULAW: info.play.encoding = AUDIO_ENCODING_ULAW; break;
    case CODE_ALAW: info.play.encoding = AUDIO_ENCODING_ALAW; break;
    case CODE_LINEAR: info.play.encoding = AUDIO_ENCODING_LINEAR; break;
    default:
      fprintf (stderr, "Help!  Bogus encoding in start_audio!\n");
      exit(1);
    }
    
    ioctl (audio_device, AUDIO_SETINFO, &info);
  }
  else
  {
    /* Was already playing --- flush data, so we start clean
     * at the new offset ...
     */
    
    ioctl (audio_device, I_SETSIG, 0);
    ioctl (audio_device, I_FLUSH, FLUSHW);
  }
  
  ioctl (audio_device, AUDIO_GETINFO, &info);

  if (portm & info.play.avail_ports)
  {
    info.play.port = portm;
    ioctl (audio_device, AUDIO_SETINFO, &info);
  }

  /* reset sample count & EOF flag */
  
  audio_reset_counts (offset); 
  
  /* Start lower half --- this begins by simulating an underrun
   * condition.  Actual playback begins when we "recover" from it.
   */
  
  init_buffers (offset);

#ifdef SVR4

  sig_act.sa_handler = feed_device;
  sigemptyset(&sig_act.sa_mask);
  sig_act.sa_flags = 0;
  sigaction(SIGPOLL,&sig_act,0);

#else

  signal (SIGPOLL, feed_device);

#endif

  ioctl (audio_device, I_SETSIG, S_OUTPUT);
}

void
audio_abort ()
{
  audio_info_t info;
  
  ioctl (audio_device, I_SETSIG, 0);
  ioctl (audio_device, I_FLUSH, FLUSHW);

  close (audio_device);
  audio_device = -1;
}


